/* Some type-punning examples.
   Most of them assume a little-endian architecture.
*/

#include <verasco.h>
#include <stdint.h>

void u8ofi32(void) {
  int32_t i = 0x12345678;
  uint8_t*u = (void*)&i;
  verasco_assert(u[0] == 0x78);
  verasco_assert(u[1] == 0x56);
  verasco_assert(u[2] == 0x34);
  verasco_assert(u[3] == 0x12);
}

void u8ofu16(void) {
  uint16_t u16 = 0x1234;
  uint8_t* u8 = (void*)&u16;
  verasco_assert(u8[1] == 0x12);
  verasco_assert(u8[0] == 0x34);
}

void u8ofs16(void) {
  int16_t s = 0x1234;
  uint8_t*u = (void*)&s;
  verasco_assert(u[1] == 0x12);
  verasco_assert(u[0] == 0x34);
}

void u8ofs8(void) {
  int8_t s = -2;
  uint8_t*u = (void*)&s;
  int32_t i = u[0]; // zero-extension
  verasco_assert(i == 254);
  i = (int8_t)(u[0]); // sign-extension
  verasco_assert(i == -2);
}

void i32ofu8(void) {
  uint8_t u8[4] = { 0x12, 0x34, 0x56, 0x78 };
  int32_t *i = (void*)u8;
  verasco_assert(i[0] == 0x78563412);
}

void i32ofi32(void) {
  int32_t i = 0x90abcdef;
  int32_t j;
  int n;
  for ( n = 0 ; n < 4 ; ++n ) {
    verasco_unroll( 4 );
    ((uint8_t*)&j)[n] = ((uint8_t*)&i)[n];
  }
  verasco_assert( j == i );
}

static
int32_t i32_from_bytes(uint8_t a, uint8_t b, uint8_t c, uint8_t d) {
  union { uint8_t x[4]; int32_t y; } v;
  v.x[0] = a;
  v.x[1] = b;
  v.x[2] = c;
  v.x[3] = d;
  return v.y;
}

void i32ofs16(void) {
  int16_t lo = 0x1234, hi = 0xfedc;
  uint8_t * plo = (void*)&lo;
  uint8_t * phi = (void*)&hi;
  verasco_assert(
    i32_from_bytes (plo[0], plo[1], phi[0], phi[1]) == 0xfedc1234
  );
}

void lctes06_fig4(void) {
  union {
    struct { uint8_t al, ah, bl, bh; } b;
    struct { uint16_t ax, bx; } w;
  } regs;

  regs.w.ax = verasco_any_int64();
  regs.w.bx = 0x1234;
  if (regs.b.ah) regs.b.bl = regs.b.al;
  else regs.b.bh = regs.b.al;
  regs.b.al = verasco_any_int64();
}

/* Inverse of the square root.
 * Public Domain
 * Taken from Matthew Robertson, “A Brief History of InvSqrt”, 2012.
 */
float Q_rsqrt(float number) {
  long i;
  float x2, y;
  const float threehalfs = 1.5f;
  x2 = number * 0.5f;
  y = number;
  i = *(long *) &y;
  i = 0x5f3759df - (i >> 1);
  y = *(float *) &i;
  y = y * (threehalfs - (x2 * y * y));
  // y = y * (threehalfs - (x2 * y * y));
  return y;
}

void test_Q_rsqrt(void) {
  float sqrt2 = 1.f / Q_rsqrt(2.f);
  float two = sqrt2 * sqrt2;
  verasco_assert ( 2.f <= two && two <= 2.005f );
}

int main(void) {
  int i = 0;
  while (1) {
    switch (i++) {
      case 0: u8ofi32(); break;
      case 1: u8ofu16(); break;
      case 2: u8ofs16(); break;
      case 3: u8ofs8(); break;
      case 4: i32ofu8(); break;
      case 5: i32ofi32(); break;
      case 6: i32ofs16(); break;
      case 7: lctes06_fig4(); break;
      /* TODO: dedicated operators in the numerical domains: bits ↔ float
      case 8: test_Q_rsqrt(); break;
      */
      default: return 0;
    }
  }
}
