(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundation,       *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

module ZArith = Z
open Camlcoq

exception Null

let rec decode_byte byte bit q =
  if bit = 0 then q
  else if byte land bit = 0 then decode_byte byte (bit lsr 1) (P.Coq_xO q)
  else decode_byte byte (bit lsr 1) (P.Coq_xI q)

let rec decode_byte_0 byte bit =
  if bit = 0 then raise Null
  else if byte land bit = 0 then decode_byte_0 byte (bit lsr 1)
  else decode_byte byte (bit lsr 1) P.Coq_xH

let decode_pos p =
  let str = ZArith.to_bits p in
  let rec loop i q =
    if i < 0 then q
    else loop (i-1) (decode_byte (int_of_char str.[i]) 128 q)
  in
  let rec loop_0 i =
    if i < 0 then raise Null
    else
      try loop (i-1) (decode_byte_0 (int_of_char str.[i]) 128)
      with Null -> loop_0 (i-1)
  in
  loop_0 (String.length str - 1)

let decode_N n = try N.Npos (decode_pos n) with Null -> N.N0

let decode_Z z =
  match ZArith.sign z with
  | 0 -> Z.Z0
  | 1 -> Z.Zpos (decode_pos z)
  | -1 -> Z.Zneg (decode_pos z)
  | _ -> assert false

let rec encode_byte i p =
  if i = 8 then (Some p, 0)
  else match p with
       | P.Coq_xH -> (None, 1)
       | P.Coq_xO p -> let (q, b) = encode_byte (i+1) p in (q, b+b)
       | P.Coq_xI p -> let (q, b) = encode_byte (i+1) p in (q, b+b+1)

let encode_pos p =
  let rec nbits = function
    | P.Coq_xH -> 1
    | P.Coq_xO p | P.Coq_xI p -> nbits p + 1
  in
  let sz = (nbits p-1)/8+1 in
  let str = Bytes.create sz in
  let rec loop p i =
    let (q, b) = encode_byte 0 p in
    Bytes.set str i (char_of_int b);
    match q with
    | None -> assert (i = sz-1)
    | Some q -> loop q (i+1)
  in
  loop p 0;
  ZArith.of_bits (Bytes.unsafe_to_string str)

let encode_N = function
  | N.N0 -> ZArith.zero
  | N.Npos p -> encode_pos p

let encode_Z = function
  | Z.Z0 -> ZArith.zero
  | Z.Zpos p -> encode_pos p
  | Z.Zneg p -> ZArith.neg (encode_pos p)

let nofz x = ZArith.max ZArith.zero x

let compare a b =
  match ZArith.compare a b with
  | -1 -> Datatypes.Lt
  | 0 -> Datatypes.Eq
  | 1 -> Datatypes.Gt
  | _ -> assert false

let div a b =
  let open ZArith in
  try fdiv a b with Division_by_zero -> zero

let modulo a b =
  let open ZArith in
  match sign b with
  | 1 -> erem a b
  | 0 -> zero
  | -1 -> neg (erem a b)
  | _ -> assert false

let quot a b =
  let open ZArith in
  try div a b with Division_by_zero -> zero

let rem a b =
  let open ZArith in
  try rem a b with Division_by_zero -> zero

(* This can raise [ZArith.Overflow] if [b] isn't in the right range.
   This is a degenerate case that should not happen. In this case,
   anyway, the equivalent Coq function takes endless time to
   compute. *)
let shiftl a b =
  match ZArith.sign b with
  | 0 -> a
  | -1 -> ZArith.shift_right a (-ZArith.to_int b)
  | 1 -> ZArith.shift_left a (ZArith.to_int b)
  | _ -> assert false

let shiftr a b =
  match ZArith.sign b with
  | 0 -> a
  | -1 -> ZArith.shift_left a (-ZArith.to_int b)
  | 1 -> ZArith.shift_right a (ZArith.to_int b)
  | _ -> assert false
